@@ -17,8 +17,11 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include <strings.h>
+
 #include <gtk/gtk.h>
 #include "htmlboxaccessible.h"
+#include "htmlboxtextaccessible.h"
 #include "a11y/htmlviewaccessible.h"
 #include <libgtkhtml/layout/htmlbox.h>
 #include <libgtkhtml/layout/htmlboxtable.h>
@@ -26,6 +29,8 @@
 #include <libgtkhtml/layout/htmlboxtablerowgroup.h>
 #include <libgtkhtml/layout/htmlboxtext.h>
 #include <libgtkhtml/layout/htmlboxinline.h>
+#include <libgtkhtml/layout/htmlboxtable.h>
+#include <libgtkhtml/layout/htmlboxtablecell.h>
 #include <libgtkhtml/dom/core/dom-element.h>
 
 static void         html_box_accessible_class_init               (HtmlBoxAccessibleClass  *klass);
@@ -110,20 +115,31 @@ html_box_accessible_new (GObject *obj)
 
 	g_return_val_if_fail (HTML_IS_BOX (obj), NULL);
 	if (HTML_IS_BOX_TEXT (obj)) {
-		HtmlBox *box = HTML_BOX (obj);
+		HtmlBoxText *text;
+		HtmlBox *box;
 
-		while (!HTML_IS_BOX_BLOCK (box)) {
-			box = box->parent;
+		text = HTML_BOX_TEXT (obj);
+		if (html_box_text_get_len (text) > 0) {
+			box = HTML_BOX (obj);
+
+			while (!HTML_IS_BOX_BLOCK (box)) {
+				box = box->parent;
+			}
+			if (!(HTML_IS_BOX_BLOCK (box) &&
+			      box->dom_node &&
+			      strcmp ((char*)box->dom_node->xmlnode->name, "p") == 0)) { 
+				atk_object = html_box_text_accessible_new (obj);
+			} else {
+				atk_object = atk_gobject_accessible_for_object (G_OBJECT (box));
+			}
+			return atk_object;
 		}
-
-		atk_object = atk_gobject_accessible_for_object (G_OBJECT (box));
-		return atk_object;
 	}
 
 	object = g_object_new (HTML_TYPE_BOX_ACCESSIBLE, NULL);
 	atk_object = ATK_OBJECT (object);
 	atk_object_initialize (atk_object, obj);
-	atk_object->role = ATK_ROLE_UNKNOWN;
+	atk_object->role = ATK_ROLE_PANEL;
 	return atk_object;
 }
 
@@ -178,7 +194,7 @@ html_box_accessible_initialize (AtkObject *obj, gpointer data)
 				g_object_set_data (G_OBJECT (parent), view_str, data);
 		}
 		parent_obj = atk_gobject_accessible_for_object (G_OBJECT (parent));
-		atk_object_set_parent (obj, parent_obj);
+		obj->accessible_parent = g_object_ref (parent_obj);
 	}
 }
 
@@ -346,7 +362,7 @@ html_box_accessible_get_parent (AtkObject *obj)
 			g_return_val_if_fail (!box->parent, NULL); 
 			g_return_val_if_fail (widget, NULL); 
 			parent = gtk_widget_get_accessible (widget);
-			atk_object_set_parent (obj, parent);
+			obj->accessible_parent = g_object_ref (parent);
 		}
 	}
 	return parent;
