@@ -31,6 +31,9 @@ static void     html_box_block_text_accessible_real_initialize (AtkObject *objec
 static gint     html_box_block_text_accessible_get_n_children  (AtkObject *obj);
 static AtkObject*   html_box_block_text_accessible_ref_child   (AtkObject *obj,
                                                                 gint      i);
+static AtkRelationSet* html_box_block_text_accessible_ref_relation_set
+                                                               (AtkObject *obj)
+;
 
 static void     html_box_block_text_accessible_text_interface_init (AtkTextIface	*iface);
 static gchar*   html_box_block_text_accessible_get_text            (AtkText             *text,
@@ -101,11 +104,14 @@ static gchar*   get_text_near_offset                               (AtkText
                                                                     gint              *start_offset,
                                                                     gint              *end_offset);
 
+extern HtmlBoxText* _html_view_get_cursor_box_text (HtmlView *view, gint *offset);
+
 static gpointer parent_class = NULL;
 
 struct _HtmlBoxBlockTextAccessiblePrivate
 {
 	GailTextUtil *textutil;
+	gint          caret_offset;
 };
 
 GType
@@ -165,6 +171,7 @@ html_box_block_text_accessible_class_init (HtmlBoxBlockAccessibleClass *klass)
 	class->initialize = html_box_block_text_accessible_real_initialize;
 	class->get_n_children = html_box_block_text_accessible_get_n_children;
 	class->ref_child = html_box_block_text_accessible_ref_child;
+	class->ref_relation_set = html_box_block_text_accessible_ref_relation_set;
 }
 
 static void
@@ -187,8 +194,6 @@ append_text (HtmlBox *root,
 
 	if (!root)
 		return;
-	if (HTML_IS_BOX_TABLE (root))
-		return;
 
 	if (HTML_IS_BOX_TEXT (root)) {
 		text = html_box_text_get_text (HTML_BOX_TEXT (root), &len);
@@ -233,73 +238,172 @@ html_box_block_text_accessible_real_initialize (AtkObject *object,
 static gint
 html_box_block_text_accessible_get_n_children (AtkObject *obj)
 {
-        AtkGObjectAccessible *atk_gobject;
-        HtmlBox *box;
-        gint n_children = 0;
-        GObject *g_obj;
-
         g_return_val_if_fail (HTML_IS_BOX_ACCESSIBLE (obj), 0);
-        atk_gobject = ATK_GOBJECT_ACCESSIBLE (obj);
-        g_obj = atk_gobject_accessible_get_object (atk_gobject);
-        if (g_obj == NULL)
-                return 0;
-
-        g_return_val_if_fail (HTML_IS_BOX (g_obj), 0);
-        box = HTML_BOX (g_obj);
-
-        if (box) {
-                HtmlBox *child;
-
-                child = box->children;
-
-                while (child) {
-                        if (HTML_IS_BOX_TABLE (child))
-				n_children++;
-                        child = child->next;
-                }
-        }
-        return n_children;
+	return 0;
 }
 
 static AtkObject*
 html_box_block_text_accessible_ref_child (AtkObject *obj,
                                           gint      i)
 {
-        AtkGObjectAccessible *atk_gobject;
-        GObject *g_obj;
-        HtmlBox *box;
-        AtkObject *atk_child = NULL;
-        gint n_children = 0;
-
         g_return_val_if_fail (HTML_IS_BOX_ACCESSIBLE (obj), NULL);
-        atk_gobject = ATK_GOBJECT_ACCESSIBLE (obj);
-        g_obj = atk_gobject_accessible_get_object (atk_gobject);
-        if (g_obj == NULL)
-                return NULL;
+	return NULL;
+}
 
-        g_return_val_if_fail (HTML_IS_BOX (g_obj), NULL);
-        box = HTML_BOX (g_obj);
+static AtkObject *
+ref_last_child (AtkObject *obj)
+{
+	AtkObject *child;
+	gint n_children;
 
-        if (box) {
-                HtmlBox *child;
+	n_children = atk_object_get_n_accessible_children (obj);
+	if (n_children > 0) {
+		child = atk_object_ref_accessible_child (obj, n_children - 1);
 
-                child = box->children;
+		while (child) {
+			n_children = atk_object_get_n_accessible_children (child);
+			if (n_children > 0) {
+				g_object_unref (child);
+				child = atk_object_ref_accessible_child (child, n_children -1);
+			} else {
+				break;
+			}
+		}
+		return child;
+	} else {
+		return NULL;
+	}
+}
 
-                while (child) {
-                        if (HTML_IS_BOX_TABLE (child)) {
-                        	if (n_children == i) {
-                               		atk_child = atk_gobject_accessible_for_object (G_OBJECT (child));
-					g_assert (HTML_IS_BOX_ACCESSIBLE (atk_child));
-					HTML_BOX_ACCESSIBLE (atk_child)->index = i;
-					g_object_ref (atk_child);
-					break;
+static AtkObject*
+ref_previous_object (AtkObject *obj)
+{
+	AtkObject *parent;
+	AtkObject *prev;
+	AtkObject *tmp;
+	gint index;
+	gint n_children;
+
+	index = atk_object_get_index_in_parent (obj);
+	parent= atk_object_get_parent (obj);
+	if (!HTML_IS_BOX_ACCESSIBLE (parent))
+		return NULL;
+
+	if (index > 0) {
+		n_children = atk_object_get_n_accessible_children (obj);
+		prev = atk_object_ref_accessible_child (parent, index - 1);
+		tmp = ref_last_child (prev);
+		if (tmp) {
+			g_object_unref (prev);
+			prev = tmp;
+		}
+	} else {
+		prev = parent;
+		while (prev) {
+			index = atk_object_get_index_in_parent (prev);
+			parent= atk_object_get_parent (prev);
+			if (!HTML_IS_BOX_ACCESSIBLE (parent))
+				return NULL;
+
+			if (index > 0) {
+				n_children = atk_object_get_n_accessible_children (obj);
+				prev = atk_object_ref_accessible_child (parent, index - 1);
+				tmp = ref_last_child (prev);
+				if (tmp) {
+					g_object_unref (prev);
+					prev = tmp;
 				}
-                        	n_children++;
-			}	
-                        child = child->next;
-                }
-        }
-        return atk_child;
+				break;
+			} else {
+				prev = parent;
+			}
+		}
+	}
+	return prev;
+}
+
+static AtkObject*
+ref_next_object (AtkObject *obj)
+{
+	AtkObject *parent;
+	AtkObject *next;
+	gint index;
+	gint n_children;
+
+	n_children = atk_object_get_n_accessible_children (obj);
+	if (n_children) {
+		return atk_object_ref_accessible_child (obj, 0);
+	}
+	parent = atk_object_get_parent (obj);
+	if (!HTML_IS_BOX_ACCESSIBLE (parent))
+	        return NULL;
+
+	index = atk_object_get_index_in_parent (obj);
+	n_children = atk_object_get_n_accessible_children (parent);
+	if (index < n_children - 1) {
+		return atk_object_ref_accessible_child (parent, index + 1);
+	} else {
+		next = parent;
+		while (next) {
+			parent = atk_object_get_parent (next);
+			if (!HTML_IS_BOX_ACCESSIBLE (parent))
+				return NULL;
+
+			index = atk_object_get_index_in_parent (next);
+			n_children = atk_object_get_n_accessible_children (parent);
+			if (index < n_children - 1) {
+				return atk_object_ref_accessible_child (parent, index + 1);
+			} else {
+				next = parent;
+			}
+		}
+		return NULL;
+	}
+}
+
+static AtkRelationSet*
+html_box_block_text_accessible_ref_relation_set (AtkObject *obj)
+{
+	AtkRelationSet *relation_set;
+	AtkObject *atk_obj;
+	AtkRelation *relation;
+	AtkObject *accessible_array[1];
+
+	relation_set = ATK_OBJECT_CLASS (parent_class)->ref_relation_set (obj);
+	if (!atk_relation_set_contains (relation_set, ATK_RELATION_FLOWS_TO)) {
+		atk_obj = ref_next_object (obj);
+		while (atk_obj) {
+			if (ATK_IS_TEXT (atk_obj))
+				break;
+			g_object_unref (atk_obj);
+			atk_obj = ref_next_object (atk_obj);
+		}
+		if (atk_obj) {
+			g_object_unref (atk_obj);
+			accessible_array [0] = atk_obj;
+			relation = atk_relation_new (accessible_array, 1, ATK_RELATION_FLOWS_TO);
+			atk_relation_set_add (relation_set, relation);
+			g_object_unref (relation);
+		}
+	}
+	if (!atk_relation_set_contains (relation_set, ATK_RELATION_FLOWS_FROM)) {
+		atk_obj = ref_previous_object (obj);
+		while (atk_obj) {
+			if (ATK_IS_TEXT (atk_obj))
+				break;
+			g_object_unref (atk_obj);
+			atk_obj = ref_previous_object (atk_obj);
+		}
+		if (atk_obj) {
+			g_object_unref (atk_obj);
+			accessible_array [0] = atk_obj;
+			relation = atk_relation_new (accessible_array, 1, ATK_RELATION_FLOWS_FROM);
+			atk_relation_set_add (relation_set, relation);
+			g_object_unref (relation);
+		}
+
+	}
+	return relation_set;
 }
 
 static void
@@ -307,7 +411,6 @@ html_box_block_text_accessible_text_interface_init (AtkTextIface *iface)
 {
  	g_return_if_fail (iface != NULL);
 
-	iface->get_text = html_box_block_text_accessible_get_text;
 	iface->get_text = html_box_block_text_accessible_get_text;
 	iface->get_text_after_offset = html_box_block_text_accessible_get_text_after_offset;
 	iface->get_text_at_offset = html_box_block_text_accessible_get_text_at_offset;
@@ -415,28 +518,78 @@ html_box_block_text_accessible_get_character_count (AtkText *text)
 	HtmlBoxBlockTextAccessible *block;
 	GtkTextBuffer *buffer;
 
-	g_return_val_if_fail (HTML_BOX_BLOCK_TEXT_ACCESSIBLE (text), NULL);
+	g_return_val_if_fail (HTML_BOX_BLOCK_TEXT_ACCESSIBLE (text), 0);
 	block = HTML_BOX_BLOCK_TEXT_ACCESSIBLE (text);
-	g_return_val_if_fail (block->priv->textutil, NULL);
+	g_return_val_if_fail (block->priv->textutil, 0);
 	buffer = block->priv->textutil->buffer;
 	return gtk_text_buffer_get_char_count (buffer);
 }
 
+static gboolean
+find_offset (HtmlBox *box, HtmlBoxText *box_text, gint *offset)
+{
+	HtmlBox *child;
+	HtmlBoxText *text;
+	gchar *char_text;
+	gint len;
+	gboolean ret;
+
+	if (HTML_IS_BOX_TEXT (box)) {
+		text = HTML_BOX_TEXT (box);
+		if (box_text == text)
+			return TRUE;
+	
+		char_text = html_box_text_get_text (text, &len);
+		len = g_utf8_strlen (char_text, len);
+		*offset += len;
+	}
+	child = box->children;
+	while (child) {
+		ret = find_offset (child, box_text, offset);
+		if (ret)
+			return ret;
+		child = child->next;
+	}		
+	return FALSE;
+}
+
 static gint
 html_box_block_text_accessible_get_caret_offset (AtkText *text)
 {
 	HtmlBoxBlockTextAccessible *block;
-	GtkTextBuffer *buffer;
-	GtkTextMark *cursor_mark;
-	GtkTextIter cursor_itr;
+        HtmlBox *box;
+        HtmlBoxText *cursor_box_text;
+        HtmlBox *cursor_box;
+        GtkWidget *widget;
+        HtmlView *view;
+        GObject *g_obj;
+        gint offset;
 
-	g_return_val_if_fail (HTML_BOX_BLOCK_TEXT_ACCESSIBLE (text), NULL);
+	g_return_val_if_fail (HTML_BOX_BLOCK_TEXT_ACCESSIBLE (text), 0);
 	block = HTML_BOX_BLOCK_TEXT_ACCESSIBLE (text);
-	g_return_val_if_fail (block->priv->textutil, NULL);
-	buffer = block->priv->textutil->buffer;
-	cursor_mark = gtk_text_buffer_get_insert (buffer);
-	gtk_text_buffer_get_iter_at_mark (buffer, &cursor_itr, cursor_mark);
-	return gtk_text_iter_get_offset (&cursor_itr);
+        g_obj = atk_gobject_accessible_get_object (ATK_GOBJECT_ACCESSIBLE (text
+));
+        if (g_obj == NULL)
+                return 0;
+
+        box = HTML_BOX (g_obj);
+        widget = html_box_accessible_get_view_widget (box);
+        view = HTML_VIEW (widget);
+        cursor_box_text = _html_view_get_cursor_box_text (view, &offset);
+	if (HTML_IS_BOX (cursor_box_text)) {
+		cursor_box = HTML_BOX (cursor_box_text);
+		while (cursor_box && !HTML_IS_BOX_BLOCK (cursor_box)) {
+			cursor_box = cursor_box->parent;
+		}
+		if (cursor_box == box) {
+			if (find_offset (box, cursor_box_text, &offset)) {
+                		block->priv->caret_offset = offset;
+			} else {
+				g_assert_not_reached ();
+			}
+		}
+	}
+        return block->priv->caret_offset;
 }
 
 static gboolean
@@ -447,13 +600,11 @@ html_box_block_text_accessible_set_caret_offset (AtkText *text,
 	GtkTextBuffer *buffer;
 	GtkTextIter pos_itr;
 
-	g_return_val_if_fail (HTML_BOX_BLOCK_TEXT_ACCESSIBLE (text), NULL);
+	g_return_val_if_fail (HTML_BOX_BLOCK_TEXT_ACCESSIBLE (text), FALSE);
 	block = HTML_BOX_BLOCK_TEXT_ACCESSIBLE (text);
-	g_return_val_if_fail (block->priv->textutil, NULL);
-	buffer = block->priv->textutil->buffer;
-	gtk_text_buffer_get_iter_at_offset (buffer,  &pos_itr, offset);
-	gtk_text_buffer_move_mark_by_name (buffer, "insert", &pos_itr);
-	return TRUE;
+	g_return_val_if_fail (block->priv->textutil, FALSE);
+	g_warning ("set_caret_offset not implemented");
+	return FALSE;
 }
 
 static gboolean
@@ -702,6 +853,7 @@ html_box_block_text_accessible_get_default_attributes (AtkText *text)
 	style = HTML_BOX_GET_STYLE (box);
 	font_spec = style->inherited->font_spec;
 	attrs = pango_attr_list_new ();
+	len = 0;
 	html_font_specification_get_all_attributes (font_spec, attrs, 0, len,
 						    HTML_VIEW (view)->magnification);
 	iter = pango_attr_list_get_iterator (attrs);
@@ -993,8 +1145,13 @@ html_box_block_text_accessible_get_selection (AtkText *text,
 			if (box_text->selection == HTML_BOX_TEXT_SELECTION_NONE)
 				break;
 			text_chars = html_box_text_get_text (box_text, NULL);
-			sel_start_index = g_utf8_strlen (text_chars, box_text->sel_start_index);
-			sel_end_index = g_utf8_strlen (text_chars, box_text->sel_end_index);
+			if (box_text->selection == HTML_BOX_TEXT_SELECTION_FULL) {
+				sel_start_index = 0;
+				sel_end_index = g_utf8_strlen (text_chars, -1);
+			} else {
+				sel_start_index = g_utf8_strlen (text_chars, box_text->sel_start_index);
+				sel_end_index = g_utf8_strlen (text_chars, box_text->sel_end_index);
+			}
 			*end_pos += sel_end_index - sel_start_index;
 			box_text = find_next_text (box, HTML_BOX (box_text));
 		}
@@ -1039,7 +1196,7 @@ html_box_block_text_accessible_add_selection (AtkText *text,
 	start_text = find_box_text_for_offset (box, &start_offset);
 	end_text = find_box_text_for_offset (box, &end_offset);
 	if (!start_text)
-		return NULL;
+		return FALSE;
 
 	box_text = start_text;
 	while (box_text) {
@@ -1116,6 +1273,102 @@ html_box_block_text_accessible_set_selection (AtkText *text,
 	return html_box_block_text_accessible_add_selection (text, start_pos, end_pos);
 }
 
+static gboolean is_text_in_line (HtmlBox       *root,
+                                 HtmlBox       *anchor,
+                                 GailOffsetType function)
+{
+	gboolean is_in_line;
+	gint anchor_y;
+	gint root_y;
+	gint delta = 1;
+
+	anchor_y = html_box_get_absolute_y (anchor);
+	root_y = html_box_get_absolute_y (root);
+	if (function == GAIL_AT_OFFSET) {
+		is_in_line = (root_y <= anchor_y + delta && 
+			      root_y >= anchor_y - delta);
+	} else if (function == GAIL_BEFORE_OFFSET) { 
+		is_in_line = (root_y + root->height <= anchor_y + delta && 
+			      root_y + root->height >= anchor_y - delta);
+	} else if (function == GAIL_AFTER_OFFSET) { 
+		is_in_line = (anchor_y + anchor->height <= root_y + delta && 
+			      anchor_y + anchor->height >= root_y - delta);
+	}
+        return is_in_line;             
+}
+
+static void
+append_text_for_line (HtmlBox       *root,
+                      HtmlBox       *anchor,
+                      GString       *content,
+                      GailOffsetType function,
+                      gint          *start,
+                      gint          *end)
+{
+	gint len;
+	HtmlBox *box;
+	gchar *text;
+
+	if (!root)
+		return;
+
+	if (HTML_IS_BOX_TEXT (root)) {
+		text = html_box_text_get_text (HTML_BOX_TEXT (root), &len);
+		if (text) {
+			if (is_text_in_line (root, anchor, function)) {
+				g_string_append_len (content, text, len);
+				if (*start == -1) {
+					*start = *end;
+				}
+			} else if (*start == -1) {
+				*end += g_utf8_strlen (text, len);
+			} else {
+				return;
+			}
+		}
+	}
+	box = root->children;
+	while (box) {
+		append_text_for_line (box, anchor, content, function, start, end);
+		box = box->next;
+	}
+}
+
+static gchar*
+get_line_near_offset (HtmlBox       *root,
+                      GailOffsetType function,
+                      gint           offset,
+                      gint          *start,
+                      gint          *end)
+{
+	HtmlBoxText *box_text;
+	GString *content;
+	gchar *line;
+	gint real_offset;
+
+	if (!root)
+		return NULL;
+
+	real_offset = offset;
+	box_text = find_box_text_for_offset (root, &real_offset);
+	if (!box_text)
+		return NULL;
+
+	*start = -1;
+	*end = 0;
+	content = g_string_new (NULL);
+	append_text_for_line (root, HTML_BOX (box_text), content, function, start, end);
+	line = g_strndup (content->str, content->len);
+	if (content->len) {
+		*end = *start + g_utf8_strlen (line, content->len);
+	} else {
+		*start = 0;
+		*end = 0;
+	}
+	g_string_free (content, TRUE);
+	return line;
+}
+ 
 static gchar*
 get_text_near_offset (AtkText          *text,
                       GailOffsetType   function,
@@ -1124,7 +1377,23 @@ get_text_near_offset (AtkText          *text,
                       gint             *start_offset,
                       gint             *end_offset)
 {
-	return gail_text_util_get_text (HTML_BOX_BLOCK_TEXT_ACCESSIBLE (text)->priv->textutil, NULL,
-					function, boundary_type, offset, 
-					start_offset, end_offset);
+	if (boundary_type == ATK_TEXT_BOUNDARY_LINE_START ||
+	    boundary_type == ATK_TEXT_BOUNDARY_LINE_END) {
+		AtkGObjectAccessible *atk_gobj;
+		GObject *g_obj;
+		HtmlBox *top_box;
+		gchar 	*text_chars;
+
+		atk_gobj = ATK_GOBJECT_ACCESSIBLE (text);
+		g_obj = atk_gobject_accessible_get_object (atk_gobj);
+		if (g_obj == NULL)
+			return NULL;
+
+		top_box = HTML_BOX (g_obj);
+      		return get_line_near_offset (top_box, function, offset, start_offset, end_offset); 
+	} else {
+		return gail_text_util_get_text (HTML_BOX_BLOCK_TEXT_ACCESSIBLE (text)->priv->textutil, NULL,
+						function, boundary_type, offset, 
+						start_offset, end_offset);
+	}
 }
